# -*- coding: utf-8 -*-

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../../'))

from hagworm.extend.logging import LogFileRotator
from hagworm.frame.tornado.base import Launcher

from routing import router
from setting import ConfigDynamic
from service.base import DataSource


def main():

    Launcher(
        router,
        ConfigDynamic.Port,
        process_num=ConfigDynamic.ProcessNum,
        on_startup=DataSource.initialize,
        on_shutdown=DataSource.release,
        debug=ConfigDynamic.Debug,
        gzip=ConfigDynamic.GZip,
        template_path=r'view',
        cookie_secret=ConfigDynamic.Secret,
        log_level=ConfigDynamic.LogLevel,
        log_file_path=ConfigDynamic.LogFilePath,
        log_file_rotation=LogFileRotator.make(ConfigDynamic.LogFileSplitSize, ConfigDynamic.LogFileSplitTime),
        log_file_retention=ConfigDynamic.LogFileBackups,
    ).start()


if __name__ == r'__main__':

    main()



# -*- coding: utf-8 -*-

from hagworm.extend.event import EventDispatcher
from hagworm.extend.asyncio.base import Utils, AsyncCirculatorForSecond, FuncWrapper, FutureWithTimeout


class EventDispatcher(EventDispatcher):
    """支持异步函数的事件总线
    """

    def _gen_observer(self):

        return FuncWrapper()


class DistributedEvent(EventDispatcher):
    """Redis实现的消息广播总线
    """

    def __init__(self, redis_pool, channel_name, channel_count):

        super().__init__()

        self._redis_pool = redis_pool

        self._channels = [f'event_bus_{Utils.md5_u32(channel_name)}_{index}' for index in range(channel_count)]

        for channel in self._channels:
            Utils.create_task(self._event_listener(channel))

    async def _event_listener(self, channel):

        async for _ in AsyncCirculatorForSecond():

            async with self._redis_pool.get_client() as cache:

                receiver, = await cache.subscribe(channel)

                Utils.log.info(f'event bus channel({channel}) receiver created')

                async for message in receiver.iter():
                    await self._event_assigner(message)

    async def _event_assigner(self, message):

        message = Utils.pickle_loads(message)

        _type = message.get(r'type', r'')
        args = message.get(r'args', [])
        kwargs = message.get(r'kwargs', {})

        if _type in self._observers:
            self._observers[_type](*args, **kwargs)

    def _gen_observer(self):

        return FuncWrapper()

    async def dispatch(self, _type, *args, **kwargs):

        channel = self._channels[Utils.md5_u32(_type) % len(self._channels)]

        message = {
            r'type': _type,
            r'args': args,
            r'kwargs': kwargs,
        }

        async with self._redis_pool.get_client() as cache:
            await cache.publish(channel, Utils.pickle_dumps(message))

    def gen_event_waiter(self, event_type, delay_time):

        return EventWaiter(self, event_type, delay_time)


class EventWaiter(FutureWithTimeout):
    """带超时的临时消息接收器
    """

    def __init__(self, dispatcher, event_type, delay_time):

        super().__init__(delay_time)

        self._dispatcher = dispatcher
        self._event_type = event_type

        self._dispatcher.add_listener(self._event_type, self._event_handler)

    def set_result(self, result):

        if self.done():
            return

        super().set_result(result)

        self._dispatcher.remove_listener(self._event_type, self._event_handler)

    def _event_handler(self, *args, **kwargs):

        if not self.done():
            self.set_result({r'args': args, r'kwargs': kwargs})



# -*- coding: utf-8 -*-

import copy

from io import BytesIO
from datetime import datetime

from xlwt import Workbook, XFStyle, Borders, Pattern


class ExcelWT(Workbook):
    """Excel生成工具
    """

    def __init__(self, name, encoding=r'utf-8', style_compression=0):

        super().__init__(encoding, style_compression)

        self._book_name = name
        self._current_sheet = None

        self._default_style = XFStyle()
        self._default_style.borders.left = Borders.THIN
        self._default_style.borders.right = Borders.THIN
        self._default_style.borders.top = Borders.THIN
        self._default_style.borders.bottom = Borders.THIN
        self._default_style.pattern.pattern = Pattern.SOLID_PATTERN
        self._default_style.pattern.pattern_fore_colour = 0x01

        self._default_title_style = copy.deepcopy(self._default_style)
        self._default_title_style.font.bold = True
        self._default_title_style.pattern.pattern_fore_colour = 0x16

    def create_sheet(self, name, titles=[]):

        sheet = self._current_sheet = self.add_sheet(name)
        style = self._default_title_style

        for index, title in enumerate(titles):
            sheet.write(0, index, title, style)
            sheet.col(index).width = 0x1200

    def add_sheet_row(self, *args):

        sheet = self._current_sheet
        style = self._default_style

        nrow = len(sheet.rows)

        for index, value in enumerate(args):
            sheet.write(nrow, index, value, style)

    def get_file(self):

        result = b''

        with BytesIO() as stream:

            self.save(stream)

            result = stream.getvalue()

        return result

    def write_request(self, request):

        filename = f"{self._book_name}.{datetime.today().strftime('%y%m%d.%H%M%S')}.xls"

        request.set_header(r'Content-Type', r'application/vnd.ms-excel')
        request.set_header(r'Content-Disposition', f'attachment;filename={filename}')

        return request.finish(self.get_file())



# -*- coding: utf-8 -*-

import signal
import socket
import asyncio

import jinja2

from tornado_jinja2 import Jinja2Loader

from tornado import version as tornado_version
from tornado.web import Application
from tornado.options import options
from tornado.process import cpu_count, fork_processes
from tornado.netutil import bind_sockets
from tornado.httpserver import HTTPServer
from tornado.platform.asyncio import AsyncIOMainLoop

from hagworm import hagworm_slogan
from hagworm import __version__ as hagworm_version
from hagworm.extend.base import Utils
from hagworm.extend.logging import DEFAULT_LOG_FILE_ROTATOR, init_logger
from hagworm.extend.interface import TaskInterface
from hagworm.extend.asyncio.base import install_uvloop
from hagworm.frame.tornado.web import LogRequestMixin


class _LauncherBase(TaskInterface):
    """启动器基类
    """

    def __init__(self, **kwargs):

        self._debug = kwargs.get(r'debug', False)

        self._process_num = kwargs.get(r'process_num', 1)

        self._on_startup = kwargs.get(r'on_startup', None)
        self._on_shutdown = kwargs.get(r'on_shutdown', None)

        self._background_service = kwargs.get(r'background_service', None)
        self._background_process = kwargs.get(r'background_process', None)

        self._process_id = 0
        self._process_num = self._process_num if self._process_num > 0 else cpu_count()

        # 后台服务任务对象
        if self._background_service is None:
            pass
        elif not isinstance(self._background_service, TaskInterface):
            raise TypeError(r'Background Service Dot Implemented Task Interface')

        # 服务进程任务对象，服务进程不监听端口
        if self._background_process is None:
            pass
        elif isinstance(self._background_process, TaskInterface):
            self._process_num += 1
        else:
            raise TypeError(r'Background Process Dot Implemented Task Interface')

        init_logger(
            kwargs.get(r'log_level', r'info').upper(),
            kwargs.get(r'log_handler', None),
            kwargs.get(r'log_file_path', None),
            kwargs.get(r'log_file_rotation', DEFAULT_LOG_FILE_ROTATOR),
            kwargs.get(r'log_file_retention', 0xff),
            self._debug
        )

        environment = Utils.environment()

        Utils.log.info(
            f'{hagworm_slogan}'
            f'hagworm {hagworm_version}\n'
            f'tornado {tornado_version}\n'
            f'python {environment["python"]}\n'
            f'system {" ".join(environment["system"])}'
        )

        install_uvloop()

        self._sockets = None
        self._event_loop = None

    def set_socket_buf(self, *, snd_buf=0, rcv_buf=0):

        for _socket in self._sockets:

            if snd_buf > 0:
                _socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, snd_buf)

            if rcv_buf > 0:
                _socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, rcv_buf)

    def start(self):

        if self._on_startup:
            self._event_loop.run_until_complete(self._on_startup())

        if self._background_service is not None:
            self._background_service.start()
            Utils.log.success(f'Background service no.{self._process_id} running...')

        if self._process_id == 0 and self._background_process is not None:
            self._background_process.start()
            Utils.log.success(f'Background process no.{self._process_id} running...')
        else:
            self._server.add_sockets(self._sockets)

        Utils.log.success(f'Startup server no.{self._process_id}')

        self._event_loop.run_forever()

        if self._on_shutdown is not None:
            self._event_loop.run_until_complete(self._on_shutdown())

    def stop(self, code=0):

        if self._background_service is not None:
            self._background_service.stop()

        if self._process_id == 0 and self._background_process is not None:
            self._background_process.stop()

        self._event_loop.stop()

        Utils.log.success(f'Shutdown server no.{self._process_id}: code.{code}')

    def is_running(self):

        return self._event_loop.is_running()


class _Application(Application):

    def log_request(self, handler):

        if isinstance(handler, LogRequestMixin):
            handler.log_request()
            super().log_request(handler)
        elif self.settings.get(r'debug') or handler.get_status() >= 400:
            super().log_request(handler)


class Launcher(_LauncherBase):
    """TornadoHttp的启动器

    用于简化和统一程序的启动操作

    """

    def __init__(self, router, port=80, *, server_settings=None, **kwargs):

        super().__init__(**kwargs)

        self._server_settings = {} if server_settings is None else server_settings

        self._app_settings = {
            r'handlers': router,
            r'debug': self._debug,
            r'gzip': kwargs.get(r'gzip', False),
        }

        if r'template_path' in kwargs:
            self._app_settings[r'template_loader'] = Jinja2Loader(
                jinja2.Environment(
                    loader=jinja2.FileSystemLoader(kwargs[r'template_path'])
                )
            )

        if r'static_path' in kwargs:
            self._app_settings[r'static_path'] = kwargs[r'static_path']

        if r'cookie_secret' in kwargs:
            self._app_settings[r'cookie_secret'] = kwargs[r'cookie_secret']

        self._sockets = bind_sockets(port)

        if self._process_num > 1:
            self._process_id = fork_processes(self._process_num)

        options.parse_command_line()

        AsyncIOMainLoop().install()

        self._event_loop = asyncio.get_event_loop()
        self._event_loop.set_debug(self._app_settings[r'debug'])

        self._event_loop.add_signal_handler(signal.SIGINT, self.stop)
        self._event_loop.add_signal_handler(signal.SIGTERM, self.stop)

        self._server = HTTPServer(_Application(**self._app_settings), **self._server_settings)



# -*- coding: utf-8 -*-

import os
import fastapi

from fastapi import __version__ as fastapi_version
from fastapi.responses import UJSONResponse

from hagworm import hagworm_label, hagworm_slogan
from hagworm import __version__ as hagworm_version
from hagworm.extend.asyncio.base import Utils
from hagworm.extend.logging import DEFAULT_LOG_FILE_ROTATOR, init_logger

from .response import ErrorResponse


DEFAULT_HEADERS = [(r'Server', hagworm_label)]


def create_fastapi(
        log_level=r'info', log_handler=None, log_file_path=None,
        log_file_rotation=DEFAULT_LOG_FILE_ROTATOR, log_file_retention=0xff,
        debug=False,
        routes=None,
        **setting
):

    init_logger(
        log_level.upper(),
        log_handler,
        log_file_path,
        log_file_rotation,
        log_file_retention,
        debug
    )

    environment = Utils.environment()

    Utils.log.info(
        f'{hagworm_slogan}'
        f'hagworm {hagworm_version}\n'
        f'fastapi {fastapi_version}\n'
        f'python {environment["python"]}\n'
        f'system {" ".join(environment["system"])}'
    )

    setting.setdefault(r'title', r'Hagworm')
    setting.setdefault(r'version', hagworm_version)

    return fastapi.FastAPI(debug=debug, routes=routes, **setting)


class Request(fastapi.Request):

    @property
    def referer(self):

        return self.headers.get(r'Referer')

    @property
    def client_ip(self):

        return self.headers.get(r'X-Read-IP', self.client.host)

    @property
    def x_forwarded_for(self):

        return Utils.split_str(self.headers.get(r'X-Forwarded-For', r''), r',')

    @property
    def content_type(self):

        return self.headers.get(r'Content-Type')

    @property
    def content_length(self):

        result = self.headers.get(r'Content-Length', r'')

        return int(result) if result.isdigit() else 0

    def get_header(self, name, default=None):

        return self.headers.get(name, default)


class APIRoute(fastapi.routing.APIRoute):

    def get_route_handler(self):

        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: fastapi.Request):

            try:
                return await original_route_handler(
                    Request(request.scope, request.receive)
                )
            except ErrorResponse as error:
                return error

        return custom_route_handler


class APIRouter(fastapi.APIRouter):
    """目录可选末尾的斜杠访问
    """

    def __init__(
            self,
            *,
            prefix=r'',
            default_response_class=UJSONResponse,
            route_class=APIRoute,
            **kwargs
    ):

        super().__init__(
            prefix=prefix,
            default_response_class=default_response_class,
            route_class=route_class,
            **kwargs
        )

    def _get_path_alias(self, path):

        _path = path.rstrip(r'/')

        if not _path:
            return [path]

        _path_split = os.path.splitext(_path)

        if _path_split[1]:
            return [_path]

        return [_path, _path + r'/']

    def api_route(self, path, *args, **kwargs):

        def _decorator(func):

            for index, _path in enumerate(self._get_path_alias(path)):

                self.add_api_route(_path, func, *args, **kwargs)

                # 兼容的URL将不会出现在docs中
                if index == 0:
                    kwargs[r'include_in_schema'] = False

            return func

        return _decorator



# -*- coding: utf-8 -*-

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.extend.asyncio.base import Utils, AsyncCirculator
from hagworm.extend.struct import NullData
from hagworm.extend.asyncio.command import Launcher, MainProcessAbstract, ChildProcessAbstract


class MainProcess(MainProcessAbstract):

    async def _run(self):

        self.set_maxsize(len(self._pids))

        self.extend(range(10))

        for _ in self._pids:
            self.append(NullData)

        async for _ in AsyncCirculator():

            stopped = self._check_pids()

            await self.recv_message()

            if stopped:
                break

    async def recv_message(self):

        while True:

            message = await self._pull_server.recv(True)

            if message is None:
                break
            elif message is NullData:
                self.pop_data()
            else:
                self.set_task_done()
                Utils.log.info(f'receive: {message}')

    async def _handle_data(self, data):

        await self._push_server.send(data)


class ChildProcess(ChildProcessAbstract):

    async def _run(self):

        await self._push_client.send(NullData)

        async for _ in AsyncCirculator():

            message = await self._pull_client.recv()

            if message is NullData:

                break

            else:

                await Utils.sleep(Utils.randint(1, 5))

                await self._push_client.send(f'{self._process_id}_{message}')


if __name__ == r'__main__':

    Launcher(subprocess=2, daemon=MainProcess()).run(ChildProcess())



# -*- coding: utf-8 -*-

import os
import aiohttp

from aiohttp.web_exceptions import HTTPBadGateway

from tornado.web import RequestHandler
from tornado.websocket import WebSocketHandler

from hagworm import hagworm_label
from hagworm.extend.struct import Result
from hagworm.extend.asyncio.base import Utils
from hagworm.extend.asyncio.net import DownloadBuffer, HTTPClientPool

from wtforms_tornado import Form


ACCESS_CONTROL_ALLOW_ORIGIN = r'*'

PROXY_IGNORE_HEADERS = (r'CONTENT-ENCODING', r'TRANSFER-ENCODING',)


def json_wraps(func):
    """json装饰器
    """

    @Utils.func_wraps(func)
    async def _wrapper(handler, *args, **kwargs):

        resp = await Utils.awaitable_wrapper(
            func(handler, *args, **kwargs)
        )

        if isinstance(resp, Result):
            return handler.write_json(resp)

    return _wrapper


class HttpBasicAuth:
    """Http基础认证装饰器
    """

    def __init__(self, realm, username, password):

        self._realm = realm
        self._username = username
        self._password = password

    def __call__(self, func):

        @Utils.func_wraps(func)
        def _wrapper(handler, *args, **kwargs):

            auth_header = handler.get_header(r'Authorization')

            try:

                if auth_header:

                    auth_info = Utils.b64_decode(auth_header.split(r' ', 2)[1])

                    if auth_info == f'{self._username}:{self._password}':
                        return func(handler, *args, **kwargs)

            except Exception as err:

                Utils.log.error(err)

            handler.set_header(
                r'WWW-Authenticate',
                f'Basic realm="{self._realm}"'
            )
            handler.set_status(401)

            return handler.finish()

        return _wrapper


class FormInjection:
    """表单注入器
    """

    def __init__(self, form_cls, err_code: int = -1):

        if not issubclass(form_cls, Form):
            raise TypeError(r'Dot Implemented Form Interface')

        self._form_cls = form_cls
        self._err_code = err_code

    def __call__(self, func):

        @Utils.func_wraps(func)
        async def _wrapper(handler: RequestBaseHandler, *args, **kwargs):

            form = self._form_cls(handler.request.arguments)

            setattr(handler, r'form', form)
            setattr(handler, r'data', form.data)

            if form.validate():

                await json_wraps(func)(handler, *args, **kwargs)

            else:

                return handler.write_json(
                    Result(self._err_code, extra=form.errors),
                    400
                )

        return _wrapper


class DebugHeader:
    """调试信息注入器
    """

    def __call__(self, func):

        @Utils.func_wraps(func)
        async def _wrapper(handler: RequestBaseHandler, *args, **kwargs):

            if self._is_debug() is True:
                self._receive_header(
                    handler.get_header(r'Debug-Data')
                )

            await func(handler, *args, **kwargs)

            if self._is_debug() is True:
                handler.set_header(
                    r'Debug-Data',
                    self._get_debug_header()
                )

        return _wrapper

    def _is_debug(self) -> bool:

        raise NotImplementedError()

    def _receive_header(self, data: str):

        raise NotImplementedError()

    def _get_debug_header(self) -> str:

        raise NotImplementedError()


class LogRequestMixin:

    def log_request(self):

        Utils.log.info(
            '\n---------- request arguments ----------\n' +
            Utils.json_encode(
                {
                    key: [Utils.basestring(val) for val in items]
                    for key, items in self.request.arguments.items()
                },
                escape_forward_slashes=False, ensure_ascii=False, indent=4
            )
        )


class _BaseHandlerMixin(Utils):
    """Handler基础工具混入类
    """

    @property
    def request_module(self):

        return f'{self.module}.{self.method}'

    @property
    def module(self):

        _class = self.__class__

        return f'{_class.__module__}.{_class.__name__}'

    @property
    def method(self):

        return self.request.method.lower()

    @property
    def version(self):

        return self.request.version.lower()

    @property
    def protocol(self):

        return self.request.protocol

    @property
    def host(self):

        return self.request.host

    @property
    def path(self):

        return self.request.path

    @property
    def query(self):

        return self.request.query

    @property
    def body(self):

        return self.request.body

    @property
    def files(self):

        return self.request.files

    @property
    def referer(self):

        return self.get_header(r'Referer', r'')

    @property
    def client_ip(self):

        return self.get_header(r'X-Read-IP', self.request.remote_ip)

    @property
    def content_type(self):

        return self.get_header(r'Content-Type', r'')

    @property
    def content_length(self):

        result = self.get_header(r'Content-Length', r'')

        return int(result) if result.isdigit() else 0

    @property
    def headers(self):

        return self.request.headers

    def get_header(self, name, default=None):
        """
        获取header数据
        """

        return self.request.headers.get(name, default)


class SocketBaseHandler(WebSocketHandler, _BaseHandlerMixin):
    """WebSocket请求处理类
    """

    @property
    def closed(self):

        return self.ws_connection is None

    def check_origin(self, origin):

        return True


class RequestBaseHandler(RequestHandler, _BaseHandlerMixin):
    """Http请求处理类
    """

    def initialize(self, **kwargs):

        setattr(self, r'_payload', kwargs)

    @property
    def payload(self):

        return getattr(self, r'_payload', None)

    @property
    def closed(self):

        return self.request.connection.stream.closed()

    def head(self, *_1, **_2):

        self.finish()

    def options(self, *_1, **_2):

        self.finish()

    async def prepare(self):

        self._parse_json_arguments()

    def set_default_headers(self):

        global ACCESS_CONTROL_ALLOW_ORIGIN

        self.set_header(r'Cache-Control', r'no-cache')
        self.set_header(r'Server', hagworm_label)

        payload = self.get_header(r'X-Client-Payload')

        if payload:
            self.set_header(r'X-Client-Payload', payload)

        origin = self.get_header(r'Origin')

        if origin:

            self.set_header(r'Access-Control-Allow-Origin', ACCESS_CONTROL_ALLOW_ORIGIN)

            method = self.get_header(r'Access-Control-Request-Method')
            if method:
                self.set_header(r'Access-Control-Allow-Methods', method)

            headers = self.get_header(r'Access-Control-Request-Headers')
            if headers:
                self.set_header(r'Access-Control-Allow-Headers', headers)

            self.set_header(r'Access-Control-Max-Age', r'86400')
            self.set_header(r'Access-Control-Allow-Credentials', r'true')

    def set_cookie(self, name, value, domain=None, expires=None, path="/", expires_days=None, **kwargs):

        if type(value) not in (str, bytes):
            value = str(value)

        return super().set_cookie(name, value, domain, expires, path, expires_days, **kwargs)

    def get_secure_cookie(self, name, value=None, max_age_days=31, min_version=None):

        result = super().get_secure_cookie(name, value, max_age_days, min_version)

        return self.basestring(result)

    def set_secure_cookie(self, name, value, expires_days=30, version=None, **kwargs):

        if type(value) not in (str, bytes):
            value = str(value)

        return super().set_secure_cookie(name, value, expires_days, version, **kwargs)

    def get_current_user(self):

        session = self.get_cookie(r'session')

        if not session:
            session = self.uuid1()
            self.set_cookie(r'session', session)

        self.current_user = session

        return session

    def compute_etag(self):

        return None

    def _parse_json_arguments(self):

        self.request.json_arguments = {}

        content_type = self.content_type

        if content_type and content_type.find(r'application/json') >= 0 and self.body:

            try:

                json_args = self.json_decode(self.body)

                if isinstance(json_args, dict):

                    self.request.json_arguments.update(json_args)

                    for key, val in self.request.json_arguments.items():

                        if not isinstance(val, str):
                            val = str(val)

                        self.request.arguments.setdefault(key, []).append(val)

            except Exception as _:

                self.log.debug(f'Invalid application/json body: {self.body}')

    def get_files(self, name):
        """
        获取files数据
        """

        result = []

        file_data = self.files.get(name, None)

        if file_data is not None:
            self.list_extend(result, file_data)

        for index in range(len(self.files)):

            file_data = self.files.get(f'{name}[{index}]', None)

            if file_data is not None:
                self.list_extend(result, file_data)

        return result

    def get_arg_str(self, name, default=r'', length=0):
        """
        获取str型输入
        """

        result = self.get_argument(name, None, True)

        if result is None:
            return default

        if not isinstance(result, str):
            result = str(result)

        if (length > 0) and (len(result) > length):
            result = result[0:length]

        return result

    def get_arg_bool(self, name, default=False):
        """
        获取bool型输入
        """

        result = self.get_argument(name, None, True)

        if result is None:
            return default

        result = self.convert_bool(result)

        return result

    def get_arg_int(self, name, default=0, min_val=None, max_val=None):
        """
        获取int型输入
        """

        result = self.get_argument(name, None, True)

        if result is None:
            return default

        result = self.convert_int(result, default)
        result = self.interval_limit(result, min_val, max_val)

        return result

    def get_arg_float(self, name, default=0.0, min_val=None, max_val=None):
        """
        获取float型输入
        """

        result = self.get_argument(name, None, True)

        if result is None:
            return default

        result = self.convert_float(result, default)
        result = self.interval_limit(result, min_val, max_val)

        return result

    def get_arg_json(self, name, default=None):
        """
        获取json型输入
        """

        result = default

        value = self.get_argument(name, None, True)

        if value is not None:
            try:
                result = self.json_decode(value)
            except Exception as _:
                self.log.debug(f'Invalid application/json argument({name}): {value}')

        return result

    def get_json_argument(self, name, default=None):

        return self.request.json_arguments.get(name, default)

    def get_json_arguments(self):

        return self.deepcopy(self.request.json_arguments)

    def get_all_arguments(self):

        result = {}

        for key in self.request.arguments.keys():
            result[key] = self.get_argument(key)

        return result

    def write_json(self, chunk, status_code=200):
        """
        输出JSON类型
        """

        self.set_header(r'Content-Type', r'application/json')

        if status_code != 200:
            self.set_status(status_code)

        result = None

        try:
            result = self.json_encode(chunk)
        except Exception as _:
            self.log.warning(f'json encode error: {chunk}')

        return self.finish(result)

    def write_png(self, chunk):
        """
        输出PNG类型
        """

        self.set_header(r'Content-Type', r'image/png')

        return self.finish(chunk)


class DownloadAgent(RequestBaseHandler, DownloadBuffer):
    """文件下载代理类
    """

    def __init__(self, *args, **kwargs):

        RequestBaseHandler.__init__(self, *args, **kwargs)
        DownloadBuffer.__init__(self)

    def on_finish(self):

        self.close()

    async def _handle_response(self, response):

        for key, val in response.headers.items():
            if key.upper() not in PROXY_IGNORE_HEADERS:
                self.set_header(key, val)

        return await DownloadBuffer._handle_response(self, response)

    async def _flush_data(self):

        while True:

            if self.closed and self.response:
                self.response.close()
                break

            chunk = self._file.read(65536)

            if chunk:
                self.write(chunk)
                await self.flush()
            elif self.finished:
                break
            else:
                await Utils.wait_frame()

    def _get_file_name(self, url):

        result = None

        try:
            result = os.path.split(self.urlparse.urlparse(url)[2])[1]
        except Exception as _:
            self.log.error(f'urlparse error: {url}')

        return result

    async def transmit(self, url, file_name=None, *, params=None, cookies=None, headers=None):

        try:

            _range = self.request.headers.get(r'Range', None)

            if _range is not None:

                if headers is None:
                    headers = {r'Range': _range}
                else:
                    headers[r'Range'] = _range

            if not file_name:
                file_name = self._get_file_name(url)

            if file_name:
                self.set_header(
                    r'Content-Disposition',
                    f'attachment;filename={file_name}'
                )

            Utils.create_task(self.fetch(url, params=params, cookies=cookies, headers=headers))

            await self._flush_data()

        except Exception as err:

            self.log.error(err)

        finally:

            self.finish()


class HTTPProxy(HTTPClientPool):
    """简易的HTTP代理类，带连接池功能
    """

    def __init__(self,
                 use_dns_cache=True, ttl_dns_cache=10,
                 limit=100, limit_per_host=0, timeout=None,
                 **kwargs
                 ):

        super().__init__(0, use_dns_cache, ttl_dns_cache, limit, limit_per_host, timeout, **kwargs)

    async def send_request(self, method: str, url: str, handler: RequestBaseHandler, **settings) -> int:

        global PROXY_IGNORE_HEADERS

        result = 0

        settings[r'data'] = handler.body
        settings[r'params'] = handler.query

        headers = dict(handler.headers)
        headers[r'Host'] = Utils.urlparse.urlparse(url).netloc
        settings[r'headers'] = headers

        settings.setdefault(r'ssl', self._ssl_context)

        try:

            async with aiohttp.ClientSession(**self._session_config) as _session:

                async with _session.request(method, url, **settings) as _response:

                    for key, val in _response.headers.items():
                        if key.upper() not in PROXY_IGNORE_HEADERS:
                            handler.set_header(key, val)

                    while True:

                        chunk = await _response.content.read(65536)

                        if chunk:
                            handler.write(chunk)
                            await handler.flush()
                        else:
                            break

                    handler.finish()

                    result = _response.status

            Utils.log.info(f'{method} {url} => status:{_response.status}')

        except aiohttp.ClientResponseError as err:

            Utils.log.warning(err)

            handler.send_error(err.status, reason=err.message)

            result = err.status

        except Exception as err:

            Utils.log.error(err)

            handler.send_error(HTTPBadGateway.status_code, reason=r'Proxy Internal Error')

            result = HTTPBadGateway.status_code

        return result

    async def get(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_GET, url, handler)

    async def options(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_OPTIONS, url, handler)

    async def head(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_HEAD, url, handler)

    async def post(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_POST, url, handler)

    async def put(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_PUT, url, handler)

    async def patch(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_PATCH, url, handler)

    async def delete(self, url: str, handler: RequestBaseHandler) -> int:

        return await self.send_request(aiohttp.hdrs.METH_DELETE, url, handler)



# -*- coding: utf-8 -*-

import qrcode

from io import BytesIO
from qrcode.image.svg import SvgPathImage


class QRCode:

    @staticmethod
    def make(data):

        code_img = BytesIO()

        qrcode.make(data).save(code_img)

        return code_img.getvalue()

    @staticmethod
    def make_svg(data):

        code_img = BytesIO()

        qrcode.make(data, image_factory=SvgPathImage).save(code_img)

        return code_img.getvalue().decode()



# -*- coding: utf-8 -*-

import ahocorasick


class TextFinder:

    def __init__(self):

        self._automaton = ahocorasick.Automaton()

    def init_words(self, words):

        self._automaton.clear()

        for val in words:
            self._automaton.add_word(val, val)

        self._automaton.make_automaton()

    def find_all(self, content):

        return tuple(self._automaton.iter_long(content))

    def replace_all(self, content, _chars=r'*'):

        words = {item[1] for item in self.find_all(content)}

        for key in words:
            content = content.replace(key, _chars * len(key))

        return content



# -*- coding: utf-8 -*-

from service.base import DataSource

from hagworm.frame.fastapi.base import APIRouter, Request
from hagworm.frame.fastapi.response import Response, ErrorResponse


router = APIRouter(default_response_class=Response)


@router.get(r'/')
async def default(request: Request):

    return request.client_ip


@router.get(r'/error')
async def error(request: Request):

    raise ErrorResponse(-1, request.client_ip, 400)


@router.get(r'/health')
async def health():

    return await DataSource().health()



# -*- coding: utf-8 -*-

from hagworm.extend.base import Utils
from hagworm.extend.struct import Configure


class _Static(Configure):

    def _init_options(self):

        ##################################################
        # MySql数据库

        self.MySqlMasterServer = self._parser.get_split_host(r'MySql', r'MySqlMasterServer')

        self.MySqlSlaveServer = self._parser.get_split_host(r'MySql', r'MySqlSlaveServer')

        self.MySqlName = self._parser.get(r'MySql', r'MySqlName')

        self.MySqlUser = self._parser.get(r'MySql', r'MySqlUser')

        self.MySqlPasswd = self._parser.get(r'MySql', r'MySqlPasswd')

        self.MySqlMasterMinConn = self._parser.getint(r'MySql', r'MySqlMasterMinConn')

        self.MySqlMasterMaxConn = self._parser.getint(r'MySql', r'MySqlMasterMaxConn')

        self.MySqlSlaveMinConn = self._parser.getint(r'MySql', r'MySqlSlaveMinConn')

        self.MySqlSlaveMaxConn = self._parser.getint(r'MySql', r'MySqlSlaveMaxConn')

        ##################################################
        # Mongo数据库

        self.MongoHost = self._parser.get_split_str(r'Mongo', r'MongoHost')

        self.MongoName = self._parser.get(r'Mongo', r'MongoName')

        self.MongoUser = self._parser.get(r'Mongo', r'MongoUser')

        self.MongoPasswd = self._parser.get(r'Mongo', r'MongoPasswd')

        self.MongoMinConn = self._parser.getint(r'Mongo', r'MongoMinConn')

        self.MongoMaxConn = self._parser.getint(r'Mongo', r'MongoMaxConn')

        ##################################################
        # 缓存

        self.RedisHost = self._parser.get_split_host(r'Redis', r'RedisHost')

        self.RedisBase = self._parser.getint(r'Redis', r'RedisBase')

        self.RedisPasswd = self._parser.get(r'Redis', r'RedisPasswd')

        self.RedisMinConn = self._parser.getint(r'Redis', r'RedisMinConn')

        self.RedisMaxConn = self._parser.getint(r'Redis', r'RedisMaxConn')

        self.RedisExpire = self._parser.getint(r'Redis', r'RedisExpire')

        self.RedisKeyPrefix = self._parser.get(r'Redis', r'RedisKeyPrefix')

        ##################################################


class _Dynamic(Configure):

    def _init_options(self):

        ##################################################
        # 基本

        self.Port = self._parser.getint(r'Base', r'Port')

        self.Debug = self._parser.getboolean(r'Base', r'Debug')

        self.GZip = self._parser.getboolean(r'Base', r'GZip')

        self.Secret = self._parser.get(r'Base', r'Secret')

        self.ProcessNum = self._parser.getint(r'Base', r'ProcessNum')

        self.ServerName = self._parser.get(r'Base', r'ServerName')

        ##################################################
        # 日志

        self.LogLevel = self._parser.get(r'Log', r'LogLevel')

        self.LogFilePath = self._parser.get(r'Log', r'LogFilePath')

        self.LogFileSplitSize = self._parser.getint(r'Log', r'LogFileSplitSize')

        self.LogFileSplitTime = self._parser.get(r'Log', r'LogFileSplitTime')

        self.LogFileBackups = self._parser.getint(r'Log', r'LogFileBackups')

        ##################################################
        # 线程池

        self.ThreadPoolMaxWorkers = self._parser.getint(r'ThreadPool', r'ThreadPoolMaxWorkers')

        ##################################################


ConfigStatic = _Static()
ConfigDynamic = _Dynamic()


cluster = Utils.getenv(r'CLUSTER', None)

if cluster is None:
    ConfigStatic.read(r'./static.conf')
    ConfigDynamic.read(r'./dynamic.conf')
else:
    ConfigStatic.read(f'./static.{cluster.lower()}.conf')
    ConfigDynamic.read(f'./dynamic.{cluster.lower()}.conf')



# -*- coding: utf-8 -*-

import os
import ssl
import aiohttp

from enum import Enum
from contextlib import contextmanager

from hagworm.extend.base import ContextManager
from hagworm.extend.asyncio.base import AsyncCirculatorForSecond
from hagworm.extend.asyncio.buffer import FileBuffer

from .base import Utils


class STATE(Enum):

    PENDING = 0x00
    FETCHING = 0x01
    SUCCESS = 0x02
    FAILURE = 0x03


DEFAULT_TIMEOUT = aiohttp.client.ClientTimeout(total=60, connect=10, sock_read=60, sock_connect=10)
DOWNLOAD_TIMEOUT = aiohttp.client.ClientTimeout(total=600, connect=10, sock_read=600, sock_connect=10)

CACERT_FILE = os.path.join(
    os.path.split(os.path.abspath(__file__))[0],
    r'../../static/cacert.pem'
)


@contextmanager
def _catch_error():
    """异常捕获

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:

        yield

    except aiohttp.ClientResponseError as err:

        if err.status < 500:
            Utils.log.warning(err)
        else:
            Utils.log.error(err)

    except Exception as err:

        Utils.log.error(err)


class _AsyncCirculator(AsyncCirculatorForSecond):

    async def _sleep(self):

        await Utils.sleep(self._current)


def _json_decoder(val, **kwargs):

    try:
        return Utils.json_decode(val, **kwargs)
    except Exception as err:
        Utils.log.warning(f'http client json decode error: {err} => {val}')


class Result(dict):

    def __init__(self, status, headers, body):

        super().__init__(status=status, headers=headers, body=body)

    def __bool__(self):

        return (self.status >= 200) and (self.status <= 299)

    @property
    def status(self):

        return self.get(r'status')

    @property
    def headers(self):

        return self.get(r'headers')

    @property
    def body(self):

        return self.get(r'body')


class _HTTPClient:
    """HTTP客户端基类
    """

    def __init__(self, retry_count=5, timeout=None, **kwargs):

        global DEFAULT_TIMEOUT

        self._ssl_context = self._create_ssl_context()

        self._retry_count = retry_count

        self._session_config = kwargs
        self._session_config[r'timeout'] = timeout if timeout is not None else DEFAULT_TIMEOUT
        self._session_config.setdefault(r'raise_for_status', True)

    async def _handle_response(self, response):

        return await response.read()

    def _create_ssl_context(self):

        global CACERT_FILE

        return ssl.create_default_context(cafile=CACERT_FILE)

    def create_timeout(self, *, total=None, connect=None, sock_read=None, sock_connect=None):
        """生成超时配置对象

        Args:
            total: 总超时时间
            connect: 从连接池中等待获取连接的超时时间
            sock_read: Socket数据接收的超时时间
            sock_connect: Socket连接的超时时间

        """

        return aiohttp.client.ClientTimeout(
            total=total, connect=connect,
            sock_read=sock_read, sock_connect=sock_connect
        )

    async def send_request(self, method, url, data=None, params=None, cookies=None, headers=None, **settings) -> Result:

        response = None

        if headers is None:
            headers = {}

        if isinstance(data, dict):
            headers.setdefault(
                r'Content-Type',
                r'application/x-www-form-urlencoded'
            )

        settings[r'data'] = data
        settings[r'params'] = params
        settings[r'cookies'] = cookies
        settings[r'headers'] = headers

        Utils.log.debug(
            r'{0} {1} => {2}'.format(
                method,
                url,
                str({key: val for key, val in settings.items() if isinstance(val, (str, list, dict))})
            )
        )

        settings.setdefault(r'ssl', self._ssl_context)

        async for times in _AsyncCirculator(max_times=self._retry_count):

            try:

                async with aiohttp.ClientSession(**self._session_config) as _session:

                    async with _session.request(method, url, **settings) as _response:

                        response = Result(
                            _response.status,
                            dict(_response.headers),
                            await self._handle_response(_response)
                        )

            except aiohttp.ClientResponseError as err:

                # 重新尝试的话，会记录异常，否则会继续抛出异常

                if err.status < 500:
                    raise err
                elif times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except aiohttp.ClientError as err:

                if times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except Exception as err:

                raise err

            else:

                Utils.log.info(f'{method} {url} => status:{response.status}')
                break

            finally:

                if times > 1:
                    Utils.log.warning(f'{method} {url} => retry:{times}')

        return response


class _HTTPTextMixin:
    """Text模式混入类
    """

    async def _handle_response(self, response):

        return await response.text()


class _HTTPJsonMixin:
    """Json模式混入类
    """

    async def _handle_response(self, response):

        return await response.json(encoding=r'utf-8', loads=_json_decoder, content_type=None)


class _HTTPTouchMixin:
    """Touch模式混入类，不接收body数据
    """

    async def _handle_response(self, response):

        return dict(response.headers)


class HTTPClient(_HTTPClient):
    """HTTP客户端，普通模式
    """

    async def get(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_GET, url, None, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def options(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_OPTIONS, url, None, params, cookies=cookies, headers=headers)

            result = resp.headers

        return result

    async def head(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_HEAD, url, None, params, cookies=cookies, headers=headers)

            result = resp.headers

        return result

    async def post(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_POST, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def put(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_PUT, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def patch(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_PATCH, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def delete(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_DELETE, url, None, params, cookies=cookies, headers=headers)

            result = resp.body

        return result


class HTTPTextClient(_HTTPTextMixin, HTTPClient):
    """HTTP客户端，Text模式
    """
    pass


class HTTPJsonClient(_HTTPJsonMixin, HTTPClient):
    """HTTP客户端，Json模式
    """
    pass


class HTTPTouchClient(_HTTPTouchMixin, HTTPClient):
    """HTTP客户端，Touch模式
    """
    pass


class HTTPClientPool(HTTPClient):
    """HTTP带连接池客户端，普通模式
    """

    def __init__(self,
                 retry_count=5, use_dns_cache=True, ttl_dns_cache=10,
                 limit=100, limit_per_host=0, timeout=None,
                 **kwargs
                 ):

        super().__init__(retry_count, timeout, **kwargs)

        self._tcp_connector = aiohttp.TCPConnector(
            use_dns_cache=use_dns_cache,
            ttl_dns_cache=ttl_dns_cache,
            ssl=self._ssl_context,
            limit=limit,
            limit_per_host=limit_per_host,
        )

        self._session_config[r'connector'] = self._tcp_connector
        self._session_config[r'connector_owner'] = False

    async def close(self):

        if not self._tcp_connector.closed:
            await self._tcp_connector.close()


class HTTPTextClientPool(_HTTPTextMixin, HTTPClientPool):
    """HTTP带连接池客户端，Text模式
    """
    pass


class HTTPJsonClientPool(_HTTPJsonMixin, HTTPClientPool):
    """HTTP带连接池客户端，Json模式
    """
    pass


class HTTPTouchClientPool(_HTTPTouchMixin, HTTPClientPool):
    """HTTP带连接池客户端，Touch模式
    """
    pass


class Downloader(_HTTPClient):
    """HTTP文件下载器
    """

    def __init__(self, file, retry_count=5, timeout=None, **kwargs):

        global DOWNLOAD_TIMEOUT

        super().__init__(
            retry_count,
            timeout if timeout is not None else DOWNLOAD_TIMEOUT,
            **kwargs
        )

        self._file = file

        self._state = STATE.PENDING

        self._response = None

    @property
    def file(self):

        return self._file

    @property
    def state(self):

        return self._state

    @property
    def finished(self):

        return self._state in (STATE.SUCCESS, STATE.FAILURE)

    @property
    def response(self):

        return self._response

    async def _handle_response(self, response):

        if self._state != STATE.PENDING:
            return

        self._state = STATE.FETCHING
        self._response = response

        with open(self._file, r'wb') as stream:

            try:

                while True:

                    chunk = await response.content.read(65536)

                    if chunk:
                        stream.write(chunk)
                    else:
                        break

            except Exception as err:

                Utils.log.error(err)

                self._state = STATE.FAILURE

            else:

                self._state = STATE.SUCCESS

        if self._state != STATE.SUCCESS and os.path.exists(self._file):
            os.remove(self._file)

    async def fetch(self, url, *, params=None, cookies=None, headers=None):

        result = False

        try:

            await self.send_request(aiohttp.hdrs.METH_GET, url, None, params, cookies=cookies, headers=headers)

            result = (self._state == STATE.SUCCESS)

        except Exception as err:

            Utils.log.error(err)

            self._state = STATE.FAILURE

        return result


class DownloadBuffer(ContextManager, Downloader):
    """HTTP文件下载器(临时文件版)
    """

    def __init__(self, timeout=None, **kwargs):

        global DOWNLOAD_TIMEOUT

        super().__init__(
            FileBuffer(),
            1,
            timeout if timeout is not None else DOWNLOAD_TIMEOUT,
            **kwargs
        )

    def _context_release(self):

        self.close()

    def close(self):

        self._file.close()

    async def _handle_response(self, response):

        if self._state != STATE.PENDING:
            return

        self._state = STATE.FETCHING
        self._response = response

        try:

            while True:

                chunk = await response.content.read(65536)

                if chunk:
                    self._file.write(chunk)
                else:
                    break

        except Exception as err:

            Utils.log.error(err)

            self._state = STATE.FAILURE

        else:

            self._state = STATE.SUCCESS



# -*- coding: utf-8 -*-

import queue
import logging

from threading import Thread
from datetime import timedelta
from loguru._file_sink import FileSink

from hagworm.extend.base import Utils


class LogFileRotator:

    @classmethod
    def make(cls, _size=500, _time=r'00:00'):

        return cls(_size, _time).should_rotate

    def __init__(self, _size, _time):

        _size = _size * (1024 ** 2)
        _time = Utils.split_int(_time, r':')

        now_time = Utils.today()

        self._size_limit = _size
        self._time_limit = now_time.replace(hour=_time[0], minute=_time[1])

        if now_time >= self._time_limit:
            self._time_limit += timedelta(days=1)

    def should_rotate(self, message, file):

        file.seek(0, 2)

        if file.tell() + len(message) > self._size_limit:
            return True

        if message.record[r'time'].timestamp() > self._time_limit.timestamp():
            self._time_limit += timedelta(days=1)
            return True

        return False


DEFAULT_LOG_FILE_ROTATOR = LogFileRotator.make()


class LogInterceptor(logging.Handler):
    """日志拦截器
    """

    def emit(self, record):

        Utils.log.opt(
            depth=6,
            exception=record.exc_info
        ).log(
            record.levelname,
            record.getMessage()
        )


DEFAULT_LOG_INTERCEPTOR = LogInterceptor()


class QueuedFileSink(FileSink):
    """日志文件队列
    """

    def __init__(self, path, *, buffer_size=0, buffer_block=False, **kwargs):

        super().__init__(path, **kwargs)

        self._buffer = queue.Queue(buffer_size)
        self._buffer_size = buffer_size
        self._buffer_block = buffer_block

        self._worker = Thread(target=self._queued_writer, daemon=True)
        self._running = True

        self._worker.start()

    def write(self, message):

        try:
            self._buffer.put(message, block=self._buffer_block)
        except queue.Full as _:
            print(f'Log queued writer overflow: {self._buffer_size}')

    def stop(self):

        self._running = False
        self._worker.join(10)

        super().stop()

    def _queued_writer(self):

        while self._running:

            try:
                message = self._buffer.get(block=True, timeout=1)
                if message:
                    super().write(message)
            except queue.Empty as _:
                pass


def init_logger(
        level, handler=None,
        file_path=None, file_rotation=DEFAULT_LOG_FILE_ROTATOR, file_retention=0xff,
        debug=False
):

    level = level.upper()

    if handler or file_path:

        Utils.log.remove()

        if handler:
            Utils.log.add(
                handler,
                level=level,
                enqueue=True,
                backtrace=debug
            )

        if file_path:
            Utils.log.add(
                QueuedFileSink(
                    Utils.path.join(file_path, f'runtime_{{time}}_{Utils.getpid()}.log'),
                    rotation=file_rotation,
                    retention=file_retention
                ),
                level=level,
                enqueue=True,
                backtrace=debug
            )

    else:

        Utils.log.level(level)

    logging.getLogger().addHandler(DEFAULT_LOG_INTERCEPTOR)



# -*- coding: utf-8 -*-


# 基础异常
class BaseError(Exception):

    def __init__(self, data=None):

        super().__init__()

        self._data = data

    @property
    def data(self):

        return self._data

    def __repr__(self):

        return repr(self._data)


# 数据库只读限制异常
class MySQLReadOnlyError(BaseError):
    pass


# 数据库客户端已销毁
class MySQLClientDestroyed(BaseError):
    pass


# 常量设置异常
class ConstError(BaseError):
    pass


# NTP校准异常
class NTPCalibrateError(BaseError):
    pass



# -*- coding: utf-8 -*-

from fastapi.responses import UJSONResponse

from hagworm.extend.struct import Result


class Response(UJSONResponse):

    def render(self, content):
        return super().render(Result(data=content))


class ErrorResponse(Exception, UJSONResponse):

    def __init__(self, error_code, content=None, status_code=200, **kwargs):

        self.error_code = error_code

        Exception.__init__(self)
        UJSONResponse.__init__(self, content, status_code, **kwargs)

    def render(self, content):
        return super().render(Result(code=self.error_code, data=content))



# -*- coding: utf-8 -*-

from service.base import DataSource
from setting import ConfigStatic, ConfigDynamic
from hagworm.extend.base import Utils
from hagworm.frame.tornado.socket import Protocol, Launcher
import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../../'))


class EchoProtocol(Protocol):

    async def connection_made(self):

        Utils.log.info(f'connection made: {self.client_address}')

    async def connection_lost(self):

        Utils.log.info(f'connection lost: {self.client_address}')

    async def data_received(self, chunk):

        await self.data_write(chunk)


def main():

    cluster = os.getenv(r'CLUSTER', None)

    if cluster is None:
        ConfigStatic.read(r'./static.conf')
        ConfigDynamic.read(r'./dynamic.conf')
    else:
        ConfigStatic.read(f'./static.{cluster.lower()}.conf')
        ConfigDynamic.read(f'./dynamic.{cluster.lower()}.conf')

    Launcher(
        EchoProtocol,
        ConfigDynamic.Port,
        on_startup=DataSource.initialize,
        on_shutdown=DataSource.release,
        debug=ConfigDynamic.Debug,
        log_level=ConfigDynamic.LogLevel,
        log_file_path=ConfigDynamic.LogFilePath,
        log_file_num_backups=ConfigDynamic.LogFileBackups,
    ).start()


if __name__ == r'__main__':
    main()



# -*- coding: utf-8 -*-

import aioredis

from aioredis.util import _NOTSET
from aioredis.commands.string import StringCommandsMixin
from aioredis.commands.transaction import Pipeline, MultiExec
from aioredis.errors import ReplyError, MaxClientsError, AuthError, ReadOnlyError

from .base import Utils, WeakContextVar, AsyncContextManager, AsyncCirculator
from .event import DistributedEvent
from .ntp import NTPClient
from .transaction import Transaction


REDIS_ERROR_RETRY_COUNT = 0x1f
REDIS_POOL_WATER_LEVEL_WARNING_LINE = 0x08


class RedisPool:
    """Redis连接管理
    """

    def __init__(self, address, password=None, *, minsize=8, maxsize=32, db=0, expire=0, key_prefix=None, **settings):

        self._pool = None
        self._expire = expire
        self._key_prefix = key_prefix

        self._settings = settings

        self._settings[r'address'] = address
        self._settings[r'password'] = password

        self._settings[r'minsize'] = minsize
        self._settings[r'maxsize'] = maxsize

        self._settings[r'db'] = db

    def __await__(self):

        self._pool = yield from aioredis.create_pool(**self._settings).__await__()

        Utils.log.info(f"Redis {self._settings[r'address']} initialized: {self._pool.size}/{self._pool.maxsize}")

        return self

    async def close(self):

        if self._pool is not None:

            self._pool.close()
            await self._pool.wait_closed()

            self._pool = None

    def get_client(self):

        client = None

        if self._pool is not None:
            client = CacheClient(self._pool, self._expire, self._key_prefix)

        return client


class RedisDelegate:
    """Redis功能组件
    """

    def __init__(self):

        self._redis_pool = None

        self._cache_client_context = WeakContextVar(f'cache_client_{Utils.uuid1()}')

    @property
    def redis_pool(self):

        return self._redis_pool

    async def async_init_redis(self, *args, **kwargs):

        self._redis_pool = await RedisPool(*args, **kwargs)

    async def async_close_redis(self):

        if self._redis_pool is not None:
            await self._redis_pool.close()
            self._redis_pool = None

    async def cache_health(self):

        result = False

        async with self.get_cache_client() as cache:
            result = bool(await cache.time())

        return result

    def get_cache_client(self, *, alone=False):

        client = None

        if alone:

            client = self._redis_pool.get_client()

        else:

            client = self._cache_client_context.get()

            if client is None:

                client = self._redis_pool.get_client()

                if client:
                    self._cache_client_context.set(client)

        return client

    def share_cache(self, cache, ckey):

        return ShareCache(cache, ckey)

    def event_dispatcher(self, channel_name, channel_count):

        return DistributedEvent(self._redis_pool, channel_name, channel_count)

    def period_counter(self, time_slice: int, key_prefix: str = r'', ntp_client: NTPClient = None):

        return PeriodCounter(self._redis_pool, time_slice, key_prefix, ntp_client)


class CacheClient(aioredis.Redis, AsyncContextManager):
    """Redis客户端对象，使用with进行上下文管理

    将连接委托给客户端对象管理，提高了整体连接的使用率

    """

    def __init__(self, pool, expire, key_prefix):

        super().__init__(None)

        self._pool = pool

        self._expire = expire

        self._key_prefix = key_prefix

    async def _init_conn(self):

        global REDIS_POOL_WATER_LEVEL_WARNING_LINE

        if self._pool_or_conn is None and self._pool:

            if (self._pool.maxsize - self._pool.size + self._pool.freesize) < REDIS_POOL_WATER_LEVEL_WARNING_LINE:
                Utils.log.warning(
                    f'Redis connection pool not enough: '
                    f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
                )
            else:
                Utils.log.debug(
                    f'Redis connection pool info: '
                    f'{self._pool.freesize}({self._pool.size}/{self._pool.maxsize})'
                )

            self._pool_or_conn = await self._pool.acquire()

    async def _close_conn(self, discard=False):

        if self._pool and self._pool_or_conn:

            connection, self._pool_or_conn = self._pool_or_conn, None

            if discard:
                connection.close()
                await connection.wait_closed()

            self._pool.release(connection)

    async def _context_release(self):

        await self._close_conn()

    async def release(self):

        await self._close_conn()

    async def _safe_execute(self, func, *args, **kwargs):

        global REDIS_ERROR_RETRY_COUNT

        result = None

        async for times in AsyncCirculator(max_times=REDIS_ERROR_RETRY_COUNT):

            try:

                await self._init_conn()

                result = await func(*args, **kwargs)

            except (ReplyError, MaxClientsError, AuthError, ReadOnlyError) as err:

                await self._close_conn(True)

                raise err

            except Exception as err:

                await self._close_conn(True)

                if times < REDIS_ERROR_RETRY_COUNT:
                    Utils.log.exception(err)
                else:
                    raise err

            else:

                break

        return result

    async def execute(self, command, *args, **kwargs):

        return await self._safe_execute(super().execute, command, *args, **kwargs)

    @staticmethod
    def _list_basestring(_list):

        return [Utils.basestring(i) for i in _list]

    @staticmethod
    def _val_encode(val):

        return Utils.pickle_dumps(val)

    @staticmethod
    def _val_decode(val):

        return Utils.pickle_loads(val)

    def _list_decode(self, _list):

        return map(
            lambda i: self._val_decode(i) if i else i,
            _list
        )

    def key(self, key, *args, **kwargs):

        if self._key_prefix:
            key = f'{self._key_prefix}_{key}'

        if not args and not kwargs:
            return key

        sign = Utils.params_sign(*args, **kwargs)

        return f'{key}_{sign}'

    def allocate_lock(self, key, expire=60):

        return MLock(self, key, expire)

    # TRANSACTION COMMANDS

    async def unwatch(self):

        return await self._safe_execute(super().unwatch)

    async def watch(self, key, *keys):

        return await self._safe_execute(super().watch, key, *keys)

    def multi_exec(self):

        return MultiExec(self._pool, aioredis.Redis)

    def pipeline(self):

        return Pipeline(self._pool, aioredis.Redis)

    # PUB/SUB COMMANDS

    async def subscribe(self, channel, *channels):

        return await self._safe_execute(super().subscribe, channel, *channels)

    async def unsubscribe(self, channel, *channels):

        return await self._safe_execute(super().unsubscribe, channel, *channels)

    async def psubscribe(self, pattern, *patterns):

        return await self._safe_execute(super().psubscribe, pattern, *patterns)

    async def punsubscribe(self, pattern, *patterns):

        return await self._safe_execute(super().punsubscribe, pattern, *patterns)

    @property
    def channels(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.pubsub_channels

    @property
    def patterns(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.pubsub_patterns

    @property
    def in_pubsub(self):

        if self._pool_or_conn is None:
            return None

        return self._pool_or_conn.in_pubsub

    # GENERIC COMMANDS

    async def delete(self, *keys):

        _keys = []

        for key in keys:

            if key.find(r'*') < 0:
                _keys.append(key)
            else:
                _keys.extend((await self.keys(key)))

        result = (await super().delete(*_keys)) if len(_keys) > 0 else 0

        return result

    def _delete(self, key, *keys):

        return super().delete(key, *keys)

    async def keys(self, pattern):

        result = await super().keys(pattern)

        if result is not None:
            result = [Utils.basestring(key) for key in result]

        return result

    def _keys(self, pattern, *, encoding=_NOTSET):

        return super().keys(pattern, encoding=encoding)

    async def randomkey(self):

        result = await super().randomkey()

        if result is not None:
            result = Utils.basestring(result)

        return result

    def _randomkey(self, *, encoding=_NOTSET):

        return super().randomkey(encoding=encoding)

    async def scan(self, cursor=0, match=None, count=None):

        result = await super().scan(cursor, match, count)

        if result is not None:
            result = (result[0], self._list_basestring(result[1]))

        return result

    def _scan(self, cursor=0, match=None, count=None):

        return super().scan(cursor, match, count)

    # STRING COMMANDS

    async def get(self, key):

        result = await super().get(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _get(self, key, *, encoding=_NOTSET):

        return super().get(key, encoding=encoding)

    async def getset(self, key, value):

        _value = self._val_encode(value)

        result = await super().getset(key, _value)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _getset(self, key, value, *, encoding=_NOTSET):

        return super().getset(key, value, encoding=encoding)

    async def mget(self, key, *keys):

        result = await super().mget(key, *keys)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _mget(self, key, *keys, encoding=_NOTSET):

        return super().mget(key, *keys, encoding=encoding)

    async def set(self, key, value, expire=0):

        _value = self._val_encode(value)
        _expire = expire if expire > 0 else self._expire

        result = await super().set(key, _value, expire=_expire)

        return result

    def _set(self, key, value, *, expire=0, pexpire=0, exist=None):

        return super().set(key, value, expire=expire, pexpire=pexpire, exist=exist)

    async def mset(self, key, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().mset(key, _value, *_pairs)

        return result

    def _mset(self, key, value, *pairs):

        return super().mset(key, value, *pairs)

    async def msetnx(self, key, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().msetnx(key, _value, *_pairs)

        return result

    def _msetnx(self, key, value, *pairs):

        return super().msetnx(key, value, *pairs)

    async def psetex(self, key, milliseconds, value):

        _value = self._val_encode(value)

        result = await super().psetex(key, milliseconds, _value)

        return result

    def _psetex(self, key, milliseconds, value):

        return super().psetex(key, milliseconds, value)

    async def setex(self, key, seconds, value):

        _value = self._val_encode(value)

        result = await super().setex(key, seconds, _value)

        return result

    def _setex(self, key, seconds, value):

        return super().setex(key, seconds, value)

    async def setnx(self, key, value):

        _value = self._val_encode(value)

        result = await super().setnx(key, _value)

        return result

    def _setnx(self, key, value):

        return super().setnx(key, value)

    # SET COMMANDS

    async def sdiff(self, key, *keys):

        result = await super().sdiff(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sdiff(self, key, *keys):

        return super().sdiff(key, *keys)

    async def sinter(self, key, *keys):

        result = await super().sinter(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sinter(self, key, *keys):

        return super().sinter(key, *keys)

    async def smembers(self, key):

        result = await super().smembers(key)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _smembers(self, key, *, encoding=_NOTSET):

        return super().smembers(key, encoding=encoding)

    async def spop(self, key):

        result = await super().spop(key)

        if result is not None:
            result = Utils.basestring(result)

        return result

    def _spop(self, key, *, encoding=_NOTSET):

        return super().spop(key, encoding=encoding)

    async def srandmember(self, key, count=1):

        result = await super().srandmember(key, count)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _srandmember(self, key, count=None):

        return super().srandmember(key, count)

    async def sunion(self, key, *keys):

        result = await super().sunion(key, *keys)

        if result is not None:
            result = self._list_basestring(result)

        return result

    def _sunion(self, key, *keys):

        return super().sunion(key, *keys)

    async def sscan(self, key, cursor=0, match=None, count=None):

        result = await super().sscan(key, cursor, match, count)

        if result is not None:
            result = (result[0], self._list_basestring(result[1]))

        return result

    def _sscan(self, key, cursor=0, match=None, count=None):

        return super().sscan(key, cursor, match, count)

    # HASH COMMANDS

    async def hget(self, key, field):

        result = await super().hget(key, field)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _hget(self, key, field, *, encoding=_NOTSET):

        return super().hget(key, field, encoding=encoding)

    async def hgetall(self, key):

        result = await super().hgetall(key)

        if result is not None:
            result = {Utils.basestring(key): self._val_decode(val) for key, val in result.items()}

        return result

    def _hgetall(self, key, *, encoding=_NOTSET):

        return super().hgetall(key, encoding=encoding)

    async def hkeys(self, key):

        result = await super().hkeys(key)

        if result is not None:
            result = [Utils.basestring(key) for key in result]

        return result

    def _hkeys(self, key, *, encoding=_NOTSET):

        return super().hkeys(key, encoding=encoding)

    async def hmget(self, key, field, *fields):

        result = await super().hmget(key, field, *fields)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _hmget(self, key, field, *fields, encoding=_NOTSET):

        return super().hmget(key, field, *fields, encoding=encoding)

    async def hmset(self, key, field, value, *pairs):

        _value = self._val_encode(value)
        _pairs = [item if index % 2 == 0 else self._val_encode(item) for index, item in enumerate(pairs)]

        result = await super().hmset(key, field, _value, *_pairs)

        return result

    def _hmset(self, key, field, value, *pairs):

        return super().hmset(key, field, value, *pairs)

    async def hmset_dict(self, key, *args, **kwargs):

        for _arg in args:
            if isinstance(_arg, dict):
                kwargs.update({key: self._val_encode(val) for key, val in _arg.items()})

        return await super().hmset_dict(key, kwargs)

    def _hmset_dict(self, key, *args, **kwargs):

        return super().hmset_dict(key, *args, **kwargs)

    async def hset(self, key, field, value):

        _value = self._val_encode(value)

        result = await super().hset(key, field, _value)

        return result

    def _hset(self, key, field, value):

        return super().hset(key, field, value)

    async def hsetnx(self, key, field, value):

        _value = self._val_encode(value)

        result = await super().hsetnx(key, field, _value)

        return result

    def _hsetnx(self, key, field, value):

        return super().hsetnx(key, field, value)

    async def hvals(self, key):

        result = await super().hvals(key)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _hvals(self, key, *, encoding=_NOTSET):

        return super().hvals(key, encoding=encoding)

    async def hscan(self, key, cursor=0, match=None, count=None):

        result = await super().hscan(key, cursor, match, count)

        if result is not None:
            result = (result[0], [(Utils.basestring(key), self._val_decode(val), ) for key, val in result[1]])

        return result

    def _hscan(self, key, cursor=0, match=None, count=None):

        return super().hscan(key, cursor, match, count)

    # LIST COMMANDS

    async def blpop(self, key, *, timeout=0):

        result = await super().blpop(key, timeout=timeout)

        if result is not None:
            result = self._val_decode(result[1])

        return result

    def _blpop(self, key, *keys, timeout=0, encoding=_NOTSET):

        return super().blpop(key, *keys, timeout=timeout, encoding=encoding)

    async def brpop(self, key, *, timeout=0):

        result = await super().brpop(key, timeout=timeout)

        if result is not None:
            result = self._val_decode(result[1])

        return result

    def _brpop(self, key, *keys, timeout=0, encoding=_NOTSET):

        return super().brpop(key, *keys, timeout=timeout, encoding=encoding)

    async def brpoplpush(self, sourcekey, destkey, timeout=0):

        result = await super().brpoplpush(sourcekey, destkey, timeout)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _brpoplpush(self, sourcekey, destkey, timeout=0, encoding=_NOTSET):

        return super().brpoplpush(sourcekey, destkey, timeout, encoding)

    async def lindex(self, key, index):

        result = await super().lindex(key, index, encoding=encoding)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _lindex(self, key, index, *, encoding=_NOTSET):

        return super().lindex(key, index, encoding=encoding)

    async def linsert(self, key, pivot, value, before=False):

        _pivot = self._val_encode(pivot)
        _value = self._val_encode(value)

        result = await super().linsert(key, _pivot, _value, before=before)

        return result

    def _linsert(self, key, pivot, value, before=False):

        return super().linsert(key, pivot, value, before=before)

    async def lpop(self, key):

        result = await super().lpop(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _lpop(self, key, *, encoding=_NOTSET):

        return super().lpop(key, encoding=encoding)

    async def lpush(self, key, value, *values):

        _value = self._val_encode(value)
        _values = [self._val_encode(val) for val in values]

        result = await super().lpush(key, _value, *_values)

        return result

    def _lpush(self, key, value, *values):

        return super().lpush(key, value, *values)

    async def lpushx(self, key, value):

        _value = self._val_encode(value)

        result = await super().lpushx(key, _value)

        return result

    def _lpushx(self, key, value):

        return super().lpushx(key, value)

    async def lrange(self, key, start, stop):

        result = await super().lrange(key, start, stop)

        if result is not None:
            result = list(self._list_decode(result))

        return result

    def _lrange(self, key, start, stop, *, encoding=_NOTSET):

        return super().lrange(key, start, stop, encoding=encoding)

    async def lrem(self, key, count, value):

        _value = self._val_encode(value)

        result = await super().lrem(key, count, _value)

        return result

    def _lrem(self, key, count, value):

        return super().lrem(key, count, value)

    async def lset(self, key, index, value):

        _value = self._val_encode(value)

        result = await super().lset(key, index, _value)

        return result

    def _lset(self, key, index, value):

        return super().lset(key, index, value)

    async def rpop(self, key):

        result = await super().rpop(key)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _rpop(self, key, *, encoding=_NOTSET):

        return super().rpop(self, key, encoding=encoding)

    async def rpoplpush(self, sourcekey, destkey):

        result = await super().rpoplpush(sourcekey, destkey)

        if result is not None:
            result = self._val_decode(result)

        return result

    def _rpoplpush(self, sourcekey, destkey, *, encoding=_NOTSET):

        return super().rpoplpush(sourcekey, destkey, encoding=encoding)

    async def rpush(self, key, value, *values):

        _value = self._val_encode(value)
        _values = [self._val_encode(val) for val in values]

        result = await super().rpush(key, _value, *_values)

        return result

    def _rpush(self, key, value, *values):

        return super().rpush(key, value, *values)

    async def rpushx(self, key, value):

        _value = self._val_encode(value)

        result = await super().rpushx(key, _value)

        return result

    def _rpushx(self, key, value):

        return super().rpushx(key, value)


class MLock(AsyncContextManager):
    """基于Redis实现的分布式锁，使用with进行上下文管理
    """

    _renew_script = '''
if redis.call("get",KEYS[1]) == ARGV[1] and redis.call("ttl",KEYS[1]) > 0 then
    return redis.call("expire",KEYS[1],ARGV[2])
else
    return 0
end
'''

    _unlock_script = '''
if redis.call("get",KEYS[1]) == ARGV[1] then
    return redis.call("del",KEYS[1])
else
    return 0
end
'''

    def __init__(self, cache, key, expire):

        self._cache = cache
        self._expire = expire

        self._lock_tag = f'process_lock_{key}'
        self._lock_val = Utils.uuid1().encode()

        self._locked = False

    @property
    def locked(self):

        return self._locked

    async def _context_release(self):

        await self.release()

    async def exists(self):

        if await self._cache.exists(self._lock_tag):
            return True
        else:
            return False

    async def acquire(self, timeout=0):

        if self._locked:

            await self.renew()

        else:

            params = {
                r'key': self._lock_tag,
                r'value': self._lock_val,
                r'expire': self._expire,
                r'exist': StringCommandsMixin.SET_IF_NOT_EXIST,
            }

            async for _ in AsyncCirculator(timeout):

                if await self._cache._set(**params):
                    self._locked = True

                if self._locked or timeout == 0:
                    break

        return self._locked

    async def wait(self, timeout=0):

        async for _ in AsyncCirculator(timeout):

            if not await self.exists():
                return True
        else:

            return False

    async def renew(self):

        if self._locked:

            if await self._cache.eval(self._renew_script, [self._lock_tag], [self._lock_val, self._expire]):
                self._locked = True
            else:
                self._locked = False

        return self._locked

    async def release(self):

        if self._locked:
            await self._cache.eval(self._unlock_script, [self._lock_tag], [self._lock_val])
            self._locked = False


class ShareCache(AsyncContextManager):
    """共享缓存，使用with进行上下文管理

    基于分布式锁实现的一个缓存共享逻辑，保证在分布式环境下，同一时刻业务逻辑只执行一次，其运行结果会通过缓存被共享

    """

    def __init__(self, cache, ckey):

        self._cache = cache
        self._ckey = ckey

        self._locker = None
        self._locked = False

        self.result = None

    async def _context_release(self):

        await self.release()

    async def get(self):

        result = await self._cache.get(self._ckey)

        if result is None:

            self._locker = self._cache.allocate_lock(self._ckey)
            self._locked = await self._locker.acquire()

            if not self._locked:
                await self._locker.wait()
                result = await self._cache.get(self._ckey)

        return result

    async def set(self, value, expire=0):

        result = await self._cache.set(self._ckey, value, expire)

        return result

    async def release(self):

        if self._locked:

            self._locked = False

            if self._locker:
                await self._locker.release()

        self._cache = self._locker = None


class PeriodCounter:

    MIN_EXPIRE = 60

    def __init__(self, cache_pool: RedisPool, time_slice: int, key_prefix: str = r'', ntp_client: NTPClient = None):

        self._cache_pool = cache_pool

        self._time_slice = time_slice
        self._key_prefix = key_prefix

        self._ntp_client = ntp_client

    def _get_key(self, key: str = None) -> str:

        timestamp = Utils.timestamp() if self._ntp_client is None else self._ntp_client.timestamp

        time_period = Utils.math.floor(timestamp / self._time_slice)

        if key is None:
            return f'{self._key_prefix}_{time_period}'
        else:
            return f'{self._key_prefix}_{key}_{time_period}'

    async def _incr(self, key: int, val: str) -> int:

        res = None

        async with self._cache_pool.get_client() as cache:
            pipeline = cache.pipeline()
            pipeline.incrby(key, val)
            pipeline.expire(key, max(self._time_slice, self.MIN_EXPIRE))
            res, _ = await pipeline.execute()

        return res

    async def _decr(self, key: int, val: str) -> int:

        res = None

        async with self._cache_pool.get_client() as cache:
            pipeline = cache.pipeline()
            pipeline.decrby(key, val)
            pipeline.expire(key, max(self._time_slice, self.MIN_EXPIRE))
            res, _ = await pipeline.execute()

        return res

    async def incr(self, val: int, key: str = None):

        _key = self._get_key(key)

        res = await self._incr(_key, val)

        return res

    async def incr_with_trx(self, val: int, key: str = None) -> (int, Transaction):

        _key = self._get_key(key)

        res = await self._incr(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._decr, _key, val)
        else:
            trx = None

        return res, trx

    async def decr(self, val: int, key: str = None) -> int:

        _key = self._get_key(key)

        res = await self._decr(_key, val)

        return res

    async def decr_with_trx(self, val: int, key: str = None) -> (int, Transaction):

        _key = self._get_key(key)

        res = await self._decr(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._incr, _key, val)
        else:
            trx = None

        return res, trx
